// partition:  a container for organizing a set of sequential integers
//             into non-overlapping groups.
// Version:    %VERSION%
// Copyright:  2013 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

package org.mattbusche.util.partition.examples;

import org.mattbusche.util.partition.*;
import java.util.Formatter;
import org.mattbusche.util.partition.examples.CrayonState.User;
import org.mattbusche.util.partition.examples.CrayonState.Quality;
import static org.mattbusche.util.partition.examples.Crayon.*;
import static org.mattbusche.util.partition.examples.CrayonState.User.*;
import static org.mattbusche.util.partition.examples.CrayonState.Quality.*;

/** Tracks the state of each Crayon via partition.
 **/
public class CrayonManager {

    private Domain<Crayon,CrayonState> crayons;
    private CrayonStateGroup state[][];

    public CrayonManager() {

        crayons = new Domain<Crayon,CrayonState>();

        // Initialize the domain by adding the 8 crayons to it.
        //
        for(Crayon c : Crayon.values())
            crayons.addEntry(c.ordinal(), c);

        // Initialize the groups by binding each group to the crayons
        // domain, and tagging each group with the CrayonState
        // it represents.
        //
        int numUser = User.values().length;
        int numQuality = Quality.values().length;

        state = new CrayonStateGroup[numUser][numQuality];

        for(User u : User.values()) {
            for(Quality q : Quality.values())
                state[u.ordinal()][q.ordinal()] =
                    new CrayonStateGroup(crayons, new CrayonState(u, q));
        }

        // Initially all crayons are sharp and in the box
        //
        state[BOX.ordinal()][SHARP.ordinal()].addAll();
    }

    public CrayonState getState(Crayon c) {
        return crayons.getValue(c.ordinal());
    }

    public User getUser(Crayon c) {
        return getState(c).user;
    }

    public Quality getQuality(Crayon c) {
        return getState(c).quality;
    }

    public void setState(Crayon c, User u, Quality q) {
        if(c != null) {
        System.out.println((new CrayonState(u, q)).toString() +
                    " <- " + getState(c) + ": " + c);
            state[u.ordinal()][q.ordinal()].addBack(c.ordinal());
        }
    }

    void setUser(Crayon c, User u) {
        if(c != null)
            setState(c, u, getQuality(c));
    }

    void setQuality(Crayon c, Quality q) {
        if(c != null)
            setState(c, getUser(c), q);
    }

    Crayon find(Quality q) {
        return find(q, BOX);
    }

    Crayon find(Quality q, User u) {
        CrayonStateGroup s = state[u.ordinal()][q.ordinal()];
        if(s.size() > 0)
            return s.peekFront().getMember();
        else
            return null;
    }

    Crayon findPreferred() {
        return findPreferred(SHARP);
    }

    Crayon findPreferred(Quality q) {
        return findPreferred(q, BOX);
    }

    Crayon findPreferred(Quality q, User u) {
        CrayonStateGroup s = state[u.ordinal()][q.ordinal()];
        if(s.size() > 0)
            return s.peekFront().getMember();
        else
            return find(q == SHARP ? DULL : SHARP, u);
    }

    Crayon findPreferred(Crayon c) {
        return findPreferred(c, SHARP);
    }

    Crayon findPreferred(Crayon c, Quality q) {
        return findPreferred(c, q, BOX);
    }

    Crayon findPreferred(Crayon c, Quality q, User u) {
        if(getUser(c) == u)
            return c;
        else
            return findPreferred(q, u);
    }

    void moveAll(User from, User to, Quality q) {
        System.out.println((new CrayonState(to, q)).toString() +
                " <- " + state[from.ordinal()][q.ordinal()]);
        state[to.ordinal()][q.ordinal()].addBack(state[from.ordinal()][q.ordinal()]);
    }

    void moveAll(User from, User to) {
        for(Quality q : Quality.values())
            moveAll(from, to, q);
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();
        for(User u : User.values())
            for(Quality q : Quality.values())
                sb.append(state[u.ordinal()][q.ordinal()] + "\n");
        return sb.toString();
    }
}
