// partition:  a container for organizing a set of sequential integers
//             into non-overlapping groups.
// Version:    0.1
// Copyright 2013 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

package org.mattbusche.util.partition;

/** <p>A resizable array of {@link Entry Entry&lt;M&gt;} objects having
 ** sequentially numbered IDs which may be partitioned across an arbitrary
 ** number of associated {@link Group}s.</p>
 **
 ** <p>Read the <a href="package-summary.html">package summary</a> for
 ** introductory information about <code>Domain</code> and its related
 ** classes.</p>
 **
 ** <p>The ID of the first <code>Entry</code> object (being the
 ** <code>Entry</code> with the lowest numbered ID in the
 ** <code>Domain</code>'s internal array) may have an arbitrary value and is
 ** tracked by <code><a href="#min">min()</a></code>.  The ID of the
 ** last <code>Entry</code> object (being the <code>Entry</code> with the
 ** highest numbered ID in the Domain's internal array) is given by <code><a
 ** href="#min">min()</a>+<a href="#size()">size()</a>-1</code>.  This
 ** internal array may be dynamically resized in either direction (either by
 ** adding entries after the end of the array which increases the maximum ID
 ** or by adding entries before the beginning of the array which decreases the
 ** minimum ID).</p>
 **
 ** <p>The methods
 ** <code><a href="#addEntry(int, M)">addEntry()</a></code>,
 ** <code><a href="#compact()">compact()</a></code>,
 ** <code><a href="#reserve(int, int)">reserve()</a></code>, and
 ** <code><a href="#resize(int, int)">resize()</a></code> (and their
 ** overloaded variants) are the operations that can result in memory
 ** reallocation.  Memory reallocation does not disturb Group membership nor
 ** does it invalidate GroupIterators.</p>
 **
 ** @author Matthew T. Busche
 ** @version 0.1
 **/
public class Domain<M,V>
{
    int      _high;
    int      _low;
    int      _size;
    int      _min;
    Object[] _node;

    /** <p>Constructs an empty <code>Domain</code>.  The <code>Domain</code>'s
     ** min property is set to 0.</p>
     **/
    public Domain()
    {
        this(0, 0);
    }

    /** <p>Constructs a <code>Domain</code> with <code>size</code> {@link
     ** Entry} objects having IDs ranging from <code>0</code> to
     ** <code>size-1</code>.  All <code>Entry</code> member fields are
     ** initialized to <code>null</code>.</p>
     **
     ** @throws PartitionException if <code>size</code> is negative.
     **/
    public Domain(int size)
    {
        this(size, 0);
    }

    /** <p>Constructs a <code>Domain</code> with <code>size</code>
     ** {@link Entry} objects having IDs ranging from <code>min</code> to
     ** <code>min+size-1</code>.  All <code>Entry</code> member fields are
     ** initialized to <code>null</p></code>.
     **
     ** @throws PartitionException if <code>size</code> is negative.
     **/
    public Domain(int size, int min)
    {
        if(size < 0)
            throw new PartitionException("Size cannot be negative.");

        _transform(min + size, min, size, min);
    }

    /** <p>Constructs a <code>Domain</code> that is a copy of <code>d</code>.
     ** Members are only copied by reference (not a deep copy or clone).  The
     ** capacity of the new <code>Domain</code> matches the size of the
     ** original (minimizing storage requirements for the copy).</p>
     **
     ** @throws NullPointerException if <code>d</code> is <code>null</code>.
     **/
    public Domain(Domain<M,V> d)
    {
        _high = d._min + d._size;
        _low = d._min;
        _size = d._size;
        _min = d._min;
        _node = new Object[_size];
        for(int i = _min; i < _min + _size; ++i)
            _node[i - _low] = new Node<M,V>(i, d.getMember(i));
    }

    /** <p>Returns the number of {@link Entry} objects in this
     ** <code>Domain</p></code>.
     **/
    public int size()
    {
        return _size;
    }

    /** <p>Returns the lowest numbered ID of this <code>Domain</code>.</p>
     **/
    public int min()
    {
        return _min;
    }

    /** <p>Returns the {@link Group} to which the {@link Entry} with given
     ** <code>id</code> belongs or <code>null</code> if that member belongs to
     ** no <code>Group</code>.</p>
     **
     ** <p>Computational Complexity:  Constant.</p>
     **
     ** @throws PartitionException if <code>id</code> is less than {@link #min}
     ** or greater than <code>{@link #min}+{@link #size}-1</code>.
     **/
    public Group<M,V> getGroup(int id)
    {
        return _getNodeChecked(id)._group;
    }

    /** <p>Returns the value of the {@link Group} to which the
     ** {@link Entry} with the given <code>id</code> belongs.  Equivalent
     ** to {@link #getGroup getGroup(id)}.{@link Group#getValue getValue()}.</p>
     **
     ** <p>Computational Complexity:  Constant.</p>
     **
     ** @throws PartitionException if <code>id</code> is less than {@link #min}
     ** or greater than <code>{@link #min}+{@link #size}-1</code>.
     **
     ** @throws NullPointerException if the <code>Entry</code> with the given
     ** <code>id</code> belongs to no <code>Group</code>.
     **/
    public V getValue(int id)
    {
        return getGroup(id)._value;
    }

    /** <p>Returns the member object associated with the given <code>id</code>.</p>
     **
     ** <p>Computational Complexity:  Constant.</p>
     **
     ** @throws PartitionException if <code>id</code> is less than {@link #min}
     ** or greater than <code>{@link #min}+{@link #size}-1</code>.
     **/
    public M getMember(int id)
    {
        return _getNodeChecked(id)._entry.member;
    }

    /** <p>Finds the {@link Entry} in this <code>Domain</code> with the given
     ** ID and sets its member field to <code>m</code>.</p>
     **
     ** <p>Computational Complexity:  Constant.</p>
     **
     ** @throws PartitionException if <code>id</code> is less than {@link #min}
     ** or greater than <code>{@link #min}+{@link #size}-1</code>.
     **/
    public void setMember(int id, M m)
    {
        _getNodeChecked(id)._entry.member = m;
    }

    /** <p>Returns the {@link Entry} object with the given <code>id</code>.</p>
     **
     ** <p>Computational Complexity:  Constant.</p>
     **
     ** @throws PartitionException if <code>id</code> is less than {@link #min}
     ** or greater than <code>{@link #min}+{@link #size}-1</code>.
     **/
    public Entry<M> getEntry(int id)
    {
        return _getNodeChecked(id)._entry;
    }

    /** <p>Equivalent to {@link #addEntry(int,Object) addEntry(id, null)}.</p>
     **/
    public void addEntry(int id)
    {
        addEntry(id, null);
    }

    /** <p>Similar to {@link #setMember setMember(id, member)} but will resize
     ** this <code>Domain</code> as neccessary to accomodate the given
     ** <code>id</code>.  If the <code>Domain</code> has size zero or if
     ** <code>id</code> is less than the current <code>min</code>, then min
     ** is set to <code>id</code>.  If the <code>Domain</code> has capacity
     ** zero, then both <code>min</code> and <code>low</code> are set to
     ** <code>id</code>.
     **/
    public void addEntry(int id, M member)
    {
        if(_size == 0)
            _transform(id+1, id, 0, id);
        else if(id < _min - 1 || id < _low)
            _transform(_high, id, (_min + _size) - (id + 1), id + 1);
        else if(id > _min + _size || id >= _high)
            _transform(id+1, _low, id - _min, _min);
        if(id >= _min && id < _min + _size)
            getEntry(id).member = member;
        else
        {
            _node[id - _low] = new Node<M,V>(id, member);
            ++_size;
            if(id < _min)
                _min = id;
        }
    }

    /** <p>Equivalent to {@link #resize(int,int) resize(newSize, min())}</p>
     **/
    public void resize(int newSize)
    {
        resize(newSize, _min);
    }

    /** <p>Resizes this <code>Domain</code> to contain <code>newSize</code>
     ** {@link Entry} objects with IDs ranging from <code>newMin</code> to
     ** <code>newMin+newSize-1</code>.</p>
     **
     ** <p>Previously existing Entry objects having IDs that fall in
     ** this newly defined range have both their member fields as well as
     ** their group memberships preserved.</p>
     **
     ** <p>Previously existing <code>Entry</code> objects having IDs that fall
     ** outside this newly defined range are removed from their
     ** <code>Group</code>s and have internal references to the
     ** <code>Entry</code> removed.</p>
     **
     ** <p>If <code>newMin</code> &lt; {@link #low} or <code>newMin+newSize
     ** &gt; {@link #high}</code> then resize will perform a reallocation to
     ** accomodate the requested <code>Entry</code> range.</p>
     **
     ** @throws PartitionException if <code>newSize</code> is negative.
     **/
    public void resize(int newSize, int newMin)
    {
        if(newSize < 0)
            throw new PartitionException("newSize cannot be negative.");
        _transform(newMin + newSize, newMin, newSize, newMin);
    }


    /** <p>Returns one plus the highest numbered ID that can be added to the
     ** <code>Domain</code> without memory reallocation.</p>
     **/
    public int high()
    {
        return _high;
    }

    /** <p>Returns the lowest numbered ID that can be added to the
     ** <code>Domain</code> without memory reallocation.</p>
     **/
    public int low()
    {
        return _low;
    }

    /** <p>Returns {@link #high high()} <code>-</code> {@link #low low()}.</p>
     **/
    public int capacity()
    {
        return _high - _low;
    }

    /** <p>Equivalent to {@link #reserve(int,int) reserve(newHigh, low())}.</p>
     **/
    public void reserve(int newHigh)
    {
        if(_low > newHigh)
            return;
        _transform(newHigh, _low, _size, _min);
    }

    /** <p>If either {@link #high} &lt; <code>newHigh</code> or {@link #low}
     ** &gt; <code>newLow</code>, then the internal array is reallocated so
     ** that <code>high</code> is increased to at least <code>newHigh</code>
     ** and <code>low</code> is reduced to at most <code>newLow</code>.  After
     ** the call, IDs from <code>newLow</code> to <code>newHigh-1</code> may
     ** subsequently added to the <code>Domain</code> without additional
     ** reallocations.</p>
     **
     ** @throws PartitionException if <code>newHigh</code> is less than
     ** <code>newLow</code>.
     **/
    public void reserve(int newHigh, int newLow)
    {
        if(newLow > newHigh)
            throw new PartitionException("newLow cannot be greater than newHigh.");
        _transform(newHigh, newLow, _size, _min);
    }

    /** <p>If the capacity of the Domain exceeds its {@link #size}, then
     ** compact reallocates internal storage reducing capacity to match the
     ** current size of the Domain.</p>
     **/
    public void compact()
    {
        if(capacity() == size())
            return;

        int newHigh = _min + _size;
        int newLow = _min;
        int newCapacity = newHigh - newLow;
        Object[] newNode = new Object[newCapacity];
        _initNodeArray(newNode, newLow, _size, _min);
        _node = newNode;
        _high = newHigh;
        _low = newLow;
    }

    @Override
    public String toString()
    {
        StringBuilder sb = new StringBuilder();
        sb.append("[");
        boolean first = true;
        int limit = min() + size();
        for(int id = min(); id < limit; ++id) {
            if(first)
                first = false;
            else
                sb.append(", ");
            Group<M,V> g = getGroup(id);
            sb.append(id);
            sb.append(":");
            if(g != null)
                sb.append(g.getValue());
            else
                sb.append("*");
        }
        sb.append("]");
        return sb.toString();
    }

    @SuppressWarnings({"unchecked"})
    private void _transform(int newHigh, int newLow, int newSize, int newMin)
    {
        if(_high == _low)
        {
            _high = newLow;
            _low  = newLow;
        }
        if(newLow < _low || newHigh > _high)
        {
            int growHigh = newHigh <= _high ? 0 : Math.max(capacity(), newHigh - _high);
            int growLow = newLow >= _low ? 0 : Math.max(capacity(), _low - newLow);
            newHigh = _high + growHigh;
            newLow = _low - growLow;
            int newCapacity = newHigh - newLow;
            Object[] newNode = new Object[newCapacity];
            _initNodeArray(newNode, newLow, newSize, newMin);
            _node = newNode;
            _low = newLow;
            _high = newHigh;
        }
        else
            _initNodeArray(_node, _low, newSize, newMin);
    }

    @SuppressWarnings("unchecked")
    private void _initNodeArray(Object[] newNode, int newLow, int newSize, int newMin)
    {
        // Throughout this method, the variable i is consistently defined as
        // an id from the domain (as opposed to an array index) and must be
        // offset by a low value to properly index into a node array.

        //   1. If there are elements at the tail of _node that are
        //      numbered greater than the highest numbered elements of
        //      newNode then these elements must be removed from their
        //      groups (since they will no longer exist in the new
        //      domain).
        //
        if(_min + _size > newMin + newSize)
        {
            int limit = Math.max(_min, newMin + newSize);
            for(int i = _min + _size - 1; i >= limit; --i)
            {
                Node<M,V> n = _getNodeUnchecked(i);
                if(n._group != null)
                    n._group._remove(n);
                _node[i-_low] = null;
            }
        }

        //   2. If there are elements at the head of _node that are
        //      numbered less than the lowest numbered elements of newNode
        //      then these elements must be removed from their groups
        //      (since they will no longer exist in the new domain).
        //
        if(_min < newMin)
        {
            for(int i = Math.min(_min + _size, newMin) - 1; i >= _min; --i)
            {
                Node<M,V> n = _getNodeUnchecked(i);
                if(n._group != null)
                    n._group._remove(n);
                _node[i-_low] = null;
            }
        }

        //   3. If there are elements at the head of newNode that are
        //      numbered less lowest numbered elements of _node then these
        //      elements must be initialized.
        //
        if(newMin < _min)
        {
            int limit = Math.min(newMin + newSize, _min);
            for(int i = newMin; i < limit; ++i)
                newNode[i - newLow] = new Node<M,V>(i);
        }

        //   4. If newNode is a newly allocated Node array then we need to copy
        //      elements with common ids from _node to newNode.
        //
        if(newNode != _node)
        {
            // Just move over the old Node objects.  Lists should be fine.
            //
            int limit = Math.min(_min + _size, newMin + newSize);
            for(int i = Math.max(_min, newMin); i < limit; ++i)
                newNode[i - newLow] = _node[i - _low];
        }

        //   5. If there are elements at the tail of newNode that are
        //      numbered greater than the highest numbered elements of
        //      _node then these elements must be initialized.
        //
        if(newMin + newSize > _min + _size)
        {
            int limit = newMin + newSize;
            for(int i = Math.max(newMin, _min + _size); i < limit; ++i)
                newNode[i - newLow] = new Node<M,V>(i);
        }

        //   6. Update the _size and _min settings.
        //
        _size = newSize;
        _min = newMin;
    }


    boolean _isInDomain(int id)
    {
        return id >= _min && id < _min + _size;
    }


    @SuppressWarnings("unchecked")
    Node<M,V> _getNodeChecked(int id)
    {
        if(!_isInDomain(id))
            throw new PartitionException("Domain<M,V>._getNode(int id = " + id +
                    "):  id not in Domain.  min=" + _min + ".  max=" + (_min + _size - 1));

        return (Node<M,V>) _node[id - _low];
    }


    @SuppressWarnings("unchecked")
    Node<M,V> _getNodeUnchecked(int id)
    {
        return (Node<M,V>) _node[id - _low];
    }
}
